#!/usr/bin/env python3

import os
import subprocess

import click
from configargparse import ArgumentParser

from . import generator
from .utils import Path, _logger

PATHS = [
    "~/src/odoo/",
    # "~/src/enterprise/",
]


def main_arguments_parser():
    def str2bool(v):
        return str(v).lower() in ('yes', 'true', 't', 'y', '1')

    parser = ArgumentParser(description='Pigeoo', default_config_files=['./pigeoo.rc', '~/.pigeoo.rc'])  # TODO: use standard config file?
    parser.add('-c', '--config', is_config_file=True, help='config file path')

    # parser
    parser.add_argument('--generate', '-g', type=str2bool, nargs='?',
                        default=True, help='Output folder for documentation.')
    parser.add_argument('--output_path', '-o', type=str, nargs='?',
                        help='Output folder for documentation.')
    parser.add_argument('--paths', '-p', type=str, nargs='?',
                        default=PATHS, help='Comma separated list of paths.')
    parser.add_argument('--local', '-l', type=str2bool, nargs='?',
                        default=True, help='If run in local mode, documentation contains links to files.')
    parser.add_argument('--modules', '-m', type=str, nargs='?',
                        default=[], help='If set, restrict the modules to their dependencies.')

    return parser


def deduplicate(seq):
   uniques = []
   [uniques.append(i) for i in seq if not uniques.count(i)]
   return uniques


def normalize_paths(paths: [Path]) -> [Path]:
    normalized_paths = []
    for path in paths:
        if os.path.split(path.rstrip("/"))[-1] == 'odoo':
            normalized_paths.append(os.path.join(path, 'addons/'))
            normalized_paths.append(os.path.join(path, 'odoo/addons/'))
        else:
            normalized_paths.append(path)
    return [os.path.expanduser(path) for path in normalized_paths]


def git_check_clean_paths(paths: [Path]):
    for path in paths:
        status = subprocess.check_output(['git', 'status'], cwd=path).decode()
        branch_name = status.splitlines()[0][len("On branch "):]
        _logger.info(f"Path {path} on branch {branch_name}.")
        if 'changes' in status.lower():
            _logger.exception("You should first clean up path:" + path + "  \n"
                              "Line numbers might be incorrect, etc.")
            if click.confirm("Do you want to clean up with `git clean -fdx`?\n", default=False):
                subprocess.check_output(['git', 'clean', '-fdx'], cwd=path)
            else:
                _logger.info("Continuing with unclean paths.")


def git_get_version(path: Path, hash=False) -> str:
    args = (['git', 'rev-parse', '--abbrev-ref', 'HEAD']
            if not hash else ['git', 'rev-parse', 'HEAD'])
    return subprocess.check_output(args, cwd=path).decode().strip()


def default_output_path(options):
    short_hashes = deduplicate(h[:8] for h in options["hashes"].values())
    return "_".join(["odoo", "_".join(options["versions"]), "_".join(short_hashes)])


def find_latest_version() -> str:
    directories = [d for d in os.listdir(os.getcwd())
                   if os.path.isdir(d) and "odoo_" in d]
    directories.sort(key=lambda x: os.path.getmtime(x))
    if not directories:
        raise Exception("No documentation found, impossible to start the server.")
    return directories[-1]


def main():
    args = main_arguments_parser().parse_args()
    if args.generate:
        # TODO: autodetect venv path, etc (project mode)
        paths = args.paths.split(',') if isinstance(args.paths, str) else PATHS
        paths = deduplicate(normalize_paths(paths))
        git_paths = [p for p in paths if "packages" not in p]
        git_check_clean_paths(git_paths)

        versions = deduplicate(git_get_version(path) for path in git_paths)
        hashes = {path: git_get_version(path, hash=True) for path in git_paths}

        options = {'local': args.local, 'hashes': hashes, 'versions': versions}
        options["git_paths"] = git_paths
        modules = args.modules.split(',') if isinstance(args.modules, str) else []
        options["modules"] = modules
        output_path = args.output_path or default_output_path(options)

        generator.main(paths, output_path, options)


if __name__ == "__main__":
    main()



import os
import shutil
from lxml import etree  # type: ignore
from lxml.builder import ElementMaker,E  # type: ignore
from lxml.html.builder import CLASS  # type: ignore
from pprint import pformat as pf

from typing import Dict

from . import formatter
from . import parser
from . import query
from .parser import InfoDepTree
from .utils import _logger, Path, file_write


STYLE = "style.css"


def index_class_name():
    return "index_class.html"


def index_module_name():
    return "index_module.html"


def format_class_tree_to_html(index_name, module_tree, class_tree, output_path: Path, options: Dict):
    class_name = ""
    for c in class_tree[0]:
     if c.get('_name'):
         class_name = c['_name']
    title = class_name
    file_name = class_name + '.html'

    body = [
        E.h1(title),
        formatter.header_to_ethtml(index_name),
        formatter.inheritance_tree_to_ethtml(module_tree, options),
        formatter.class_tree_to_ethtml(class_tree, options),
    ]
    return html_write(title, body, file_name, output_path)


def html_write(title: str, body, file_name: str, output_path: Path):
    body = E.body(*body)
    content = html_generate(title, body)
    file_write(content, os.path.join(output_path, file_name))
    return title, file_name


def format_module_tree_to_html(index_name: str, module: str, module_tree: InfoDepTree, output_path: Path, options):
    body = [
        E.h1(module),
        formatter.header_to_ethtml(index_name),
        formatter.inheritance_tree_to_ethtml(module_tree, options),
    ]
    return html_write(module, body, module + '.html', output_path)


def html_generate_index(title:str, name: str, file_names, output_path: Path):
    index_file = os.path.join(output_path, name)
    file_names.sort()

    e = E.div(CLASS("blocky indnt"))
    body = E.div(CLASS("blocky"), E.h1(title), e)
    for class_name, file_name in file_names:
        e.append(E.div(CLASS("blocky"), formatter.html_link(file_name, class_name)))

    content = html_generate(title, body)

    return file_write(content, index_file)


def html_generate(title, body):
    M = ElementMaker()
    html = M.html(
        E.head(
            E.meta(charset="utf-8"),
            E.link(rel="stylesheet", href=STYLE, type="text/css"),
            E.title(title),
        ),
        body,
        lang="en",
    )
    result = etree.tostring(html,
                            doctype='<!DOCTYPE html>',
                            encoding='unicode',
                            method='xml',
                            pretty_print=True)
    return result


def html_generate_doc(class_list, output_path, options:Dict):
    index_name = index_class_name()
    title = "Odoo Class Index"
    file_names = [
        format_class_tree_to_html(index_name, module_tree, class_tree, output_path, options)
        for module_tree, class_tree in class_list
    ]
    html_generate_index(title, index_name, file_names, output_path)
    return index_name


def html_generate_modules(module_list, output_path, options:Dict):
    index_name = index_module_name()
    title = "Odoo Module Index"
    file_names = [
        format_module_tree_to_html(index_name, module, dep_tree, output_path, options)
        for module, dep_tree in module_list.items()
    ]
    html_generate_index(title, index_name, file_names, output_path)
    return index_name


def main_generate_doc(paths: [Path], all_module_deps, output_path: Path, options:Dict):
    all_modules = parser.modules_from_paths(paths, all_module_deps)
    all_models_files = parser.models_files_from_modules(all_modules)
    all_model_dicts = []
    for model_file in all_models_files:
        try:
            all_model_dicts.extend(parser.parse_model_file(
                parser.module_name_from_path(model_file), model_file))
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            _logger.exception("Parsing %s:" % model_file)

    file_write(pf(all_model_dicts), os.path.join(output_path, "all_classes.py"))

    all_class_names = {c['_name'] for c in all_model_dicts if '_name' in c}

    all_class_trees = []
    for name in all_class_names:
        try:
            c = parser.class_tree(name, all_model_dicts, paths, options, all_module_deps)
            all_class_trees.append(c)
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            _logger.exception("Processing %s:" % name)

    return html_generate_doc(all_class_trees, output_path, options)


def generate_module_deps(paths: [Path], options:Dict):
    all_modules = parser.modules_from_paths(paths)

    all_module_deps = {}
    for m in all_modules:
        deps = parser.module_dependencies_tree(m, paths)
        ideps = parser.invert_dependencies(m, deps, paths)
        n = parser.module_name_from_path(m)
        infos = parser.dep_tree_enrich(ideps, paths, options)

        all_module_deps.update({n: infos})
    return all_module_deps

def main_generate_module_deps(all_module_deps, output_path: Path, options: Dict) -> str:
    file_write(pf(all_module_deps), os.path.join(output_path, "all_modules.py"))
    return html_generate_modules(all_module_deps, output_path, options)


def filter_modules(all_module_deps, options: Dict):
    result = all_module_deps
    if options["modules"]:
        filter = options["modules"]
        depends = query.module_m_depends_on_n
        result = {
            m: all_module_deps[m] for m in all_module_deps if
            m in filter or any(depends(f, m, all_module_deps) for f in filter)
        }
    return result


def copy_stylesheet(output_path):
    source = os.path.dirname(os.path.realpath(__file__))
    stylesheet = os.path.join(source, "static/", STYLE)
    shutil.copyfile(stylesheet, os.path.join(output_path, STYLE))


def main(paths, output_path, options):
    _logger.info("Starting documentation for " + output_path)
    os.makedirs(output_path, mode=0o777, exist_ok=True)

    all_module_deps = generate_module_deps(paths, options)
    all_module_deps = filter_modules(all_module_deps, options)
    main_generate_module_deps(all_module_deps, output_path, options)
    main_generate_doc(paths, all_module_deps, output_path, options)

    copy_stylesheet(output_path)
    file_write(pf(options), os.path.join(output_path, "options.py"))

    _logger.info("Documentation has been generated.")



def get_class_name(c):
    if c.get('_name'):
        return c.get('_name')
    if c.get('_inherit') and isinstance(c.get('_inherit'), list):
        return c.get('_inherit')[0]
    return c.get('_inherit')


def get_class(class_name, class_list):
    return [c for c in class_list if get_class_name(c) == class_name]


def get_depending_modules(mod_name, all_modules) -> [str]:  # -> [module_names]
    return [m for m in all_modules if module_m_depends_on_n(m, mod_name, all_modules)]


def module_m_depends_on_n(m, n, all_modules) -> bool:
    # we only need to look at level k + 1 if module_name has dependency depth k
    if m == n:
        return False
    k = len(all_modules[n])
    result = False
    if len(all_modules[m]) > k:
        result = any(n == md["name"] for md in all_modules[m][k])
    return result

def get_functions(name, function_list):
    functions = {key[1]: function_list[key] for key in function_list if key[0] == name}
    return functions


def get_all_functions(class_list):
    return get_all_entities("functions", class_list)


def get_all_fields(class_list):
    return get_all_entities("fields", class_list)


def get_all_entities(entity, class_list):
    all_entities = {}
    for c in class_list:
        class_name = get_class_name(c)
        if class_name:
            for function in c[entity]:
                key = (function, class_name)
                all_entities.setdefault(key, {})
                all_entities[key][c["module"]] = c[entity][function]
    return all_entities



from typing import Dict

from lxml.builder import E
from lxml.html.builder import CLASS  # type: ignore

from . import parser


WEB_ICON = "🌐"


def html_link(link: str, name: str=""):
    return E.a(name or link, href=link)


def header_to_ethtml(name: str):
    e = E.div(CLASS("blocky"))
    e.append(html_link(name, "Return to Index"))
    return e


def class_to_ethtml(odoo_class, options: Dict):
    root = E.div(CLASS("flowy maxthird"))

    github_link = parser.web_link(odoo_class['file'], options)

    details = E.details()
    if options['local']:
        span = [html_link(odoo_class['full path'], odoo_class['module'])]
    else:
        span = [odoo_class['module']]
    if github_link:
        span.append(html_link(github_link, WEB_ICON))
    summary = E.summary(E.span(*span))
    e = E.div()
    root.append(details)
    details.append(summary)
    details.append(e)

    for attribute in parser.special_attributes:
        a = odoo_class.get(attribute, False)
        if a:
            e.append(E.div(attribute + ': '+ str(a), CLASS("blocky")))

    fields = odoo_class['fields']
    if len(fields):
        e_d = E.details(E.summary("Fields"))
        e_fields = E.div(CLASS("flowy-row"), e_d)
        for field in fields:
            div = []
            if github_link:
                line_no = str(fields[field]['lineno'])
                div.append(html_link(github_link + "#L" + line_no, WEB_ICON))
            div += [field + ": " + fields[field]['type'], CLASS("indnt")]
            f = E.div(*div)
            e_d.append(f)
        e.append(e_fields)

    functions = odoo_class['functions']
    if len(functions):
        e_d = E.details(E.summary("Functions"))
        e_functions = E.div(CLASS("flowy-row"), e_d)
        for function in functions:
            div = []
            if github_link:
                line_no = str(functions[function]['lineno'])
                div.append(html_link(github_link + "#L" + line_no, WEB_ICON))
            div += [function, CLASS("indnt")]
            f = E.div(*div)
            e_d.append(f)
        e.append(e_functions)

    return root


def class_tree_to_ethtml(class_tree, options):
    e = E.div(CLASS("blocky"), E.h2("Class tree"))
    for level in class_tree:
        l = E.div(CLASS("flowy-row f_c"))
        for odoo_class in level:
            l.append(class_to_ethtml(odoo_class, options))
        e.append(l)
    return e


def inheritance_tree_to_ethtml(module_tree, options: Dict):
    e = E.div(CLASS("growy"))
    for level in module_tree:
        l = E.div(CLASS("flowy-row f_c"))
        for module in level:
            if options["local"]:
                span = [html_link(module["path"], module["name"])]
            else:
                span = [module["name"]]
            if module["link"]:
                span.append(html_link(module["link"], ' 🌐'))
            l.append(E.div(E.span(*span), CLASS("flowy f_c")))
        e.append(l)
    return E.div(CLASS("blocky"), E.h2("Inheritance tree"), e)


def functions_to_ethtml(name, fdict, options: Dict):
    e = E.div(CLASS("growy"))
    for model in fdict:
        ul = E.ul(*[E.li(m) for m in list(fdict[model].keys())])
        l = E.div(CLASS("flowy-row f_c"), E.h4(model), ul)
        e.append(l)
    return E.div(CLASS("blocky"), E.h2(name), e)
